/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

// gcx.c: main file for camera control program 
// $Revision: 1.27 $
// $Date: 2004/02/24 01:50:59 $

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/io.h>
#include <sys/time.h>
#include <unistd.h>
#include <getopt.h>
#include <pwd.h>
#include <errno.h>

#include "gcx.h"
#include "x11ops.h"
#include "catalogs.h"
#include "gui.h"
#include "params.h"
#include "filegui.h"
#include "obslist.h"
#include "helpmsg.h"
#include "recipy.h"
#include "reduce.h"

static void show_usage(void) {
	info_printf("%s", help_usage_page);
//	info_printf("\nFrame stacking options\n");
//	info_printf("-o <ofile>   Set output file name (for -c, -s, -a, -p) \n");
//	info_printf("-c <method>  combine files in the file list\n");
//	info_printf("             using the specified method. Valid methods\n");
//	info_printf("             are: mm(edian), av(erage), me(dian)\n");
//	info_printf("-s <method>  Stack (align and combine) files in the file\n");
//	info_printf("             list using same methods as -c\n");
//	info_printf("-a           align files in-place\n");
//	info_printf("-w <value>   Max FWHM accepted when auto-stacking frames [5.0]\n");
//	info_printf("\nFrame processing options\n");
//	info_printf("-d <dkfile>  Substract <dkfile> from files\n");
//	info_printf("-f <fffile>  Flatfield files using <fffile> (maintain avg flux)\n");
//	info_printf("-b <bfile>   Fix bad pixels in fits files using <bfile>\n");
//	info_printf("-m <mult>    Multiply files by the scalar <mult>\n");
//	info_printf("Any of -d, -f, -b and -m can be invoked at the same time; in this\n");
//	info_printf("    case, the order of processing is dark, flat, badpix, multiply\n");
//	info_printf("    These options will cause the frame stack options to be ingnored\n");
}

static void help_all(void) {
	printf("%s\n\n", help_usage_page);
	printf("%s\n\n", help_bindings_page);
	printf("%s\n\n", help_obscmd_page);
	printf("%s\n\n", help_rep_conv_page);
}

static int load_par_file(char *fn, GcxPar p)
{
	FILE *fp;
	int ret;

	fp = fopen(fn, "r");
	if (fp == NULL) {
		err_printf("load_par_file: cannot open %s\n", fn);
		return -1;
	}

	ret = fscan_params(fp, p);
	fclose(fp);
	return ret;
}

static int save_par_file(char *fn, GcxPar p)
{
	FILE *fp;
	fp = fopen(fn, "w");
	if (fp == NULL) {
		err_printf("save_par_file: cannot open %s\n", fn);
		return -1;
	}
	fprint_params(fp, p);
	fclose(fp);
	return 0;
}

static int load_params_rc(void)
{
	uid_t my_uid;
	struct passwd *passwd;
	char *rcname;
	int ret;

	my_uid = getuid();
	passwd = getpwuid(my_uid);
	if (passwd == NULL) {
		err_printf("load_params_rc: cannot determine home directoy\n");
		return -1;
	}
	asprintf(&rcname, "%s/.gcxrc", passwd->pw_dir);
	ret = load_par_file(rcname, PAR_NULL);
	free(rcname);
	return ret;
}

int save_params_rc(void)
{
	uid_t my_uid;
	struct passwd *passwd;
	char *rcname;
	int ret;

	my_uid = getuid();
	passwd = getpwuid(my_uid);
	if (passwd == NULL) {
		err_printf("save_params_rc: cannot determine home directoy\n");
		return -1;
	}
	asprintf(&rcname, "%s/.gcxrc", passwd->pw_dir);
	ret = save_par_file(rcname, PAR_NULL);
	free(rcname);
	return ret;
}

static void recipy_file_convert(char *rcf, char *outf)
{
	FILE *infp = NULL, *outfp = NULL;
	int nc;
	if (rcf[0] != 0) {
		infp = fopen(rcf, "r");
		if (infp == NULL) {
			err_printf("Cannot open file %s for reading\n%s\n", 
				   rcf, strerror(errno));
			exit(1);
		}
	}
	if (outf[0] != 0) {
		outfp = fopen(outf, "w");
		if (infp == NULL) {
			err_printf("Cannot open file %s for writing\n%s\n", 
				   outf, strerror(errno));
			fclose(infp);
			exit(1);
		}
	}
	if (outfp && infp)
		nc = convert_recipy(infp, outfp);
	else if (outfp)
		nc = convert_recipy(stdin, outfp);
	else if (infp)
		nc = convert_recipy(infp, stdout);
	else
		nc = convert_recipy(stdin, stdout);

	if (nc < 0) {
		err_printf("Error converting recipy\n");
		if (infp)
			fclose(infp);
		if (outfp)
			fclose(outfp);
		exit(1);
	} else {
		if (infp)
			fclose(infp);
		if (outfp) {
			info_printf("%d star(s) written\n", nc);
			fclose(outfp);
		}
		exit (0);
	}
}

static void report_convert(char *rcf, char *outf)
{
	FILE *infp = NULL, *outfp = NULL;

	if (rcf[0] != 0) {
		infp = fopen(rcf, "r");
		if (infp == NULL) {
			err_printf("Cannot open file %s for reading\n%s\n", 
				   rcf, strerror(errno));
			exit(1);
		}
	}
	if (outf[0] != 0) {
		outfp = fopen(outf, "w");
		if (infp == NULL) {
			err_printf("Cannot open file %s for writing\n%s\n", 
				   outf, strerror(errno));
			if (infp)
				fclose(infp);
			exit(1);
		}
	}
	if (outfp && infp)
		report_to_table(infp, outfp, NULL);
	else if (infp)
		report_to_table(infp, stdout, NULL);
	else if (outfp)
		report_to_table(stdin, outfp, NULL);
	else
		report_to_table(stdin, stdout, NULL);
	if (infp)
		fclose(infp);
	if (outfp) 
		fclose(outfp);
	exit (0);
}


static void landolt_file_convert(char *rcf, char *outf)
{
	FILE *infp, *outfp = NULL;
	int nc;
	infp = fopen(rcf, "r");
	if (infp == NULL) {
		err_printf("Cannot open file %s for reading\n%s\n", 
			   rcf, strerror(errno));
		exit(1);
	}
	if (outf[0] != 0) {
		outfp = fopen(outf, "w");
		if (infp == NULL) {
			err_printf("Cannot open file %s for writing\n%s\n", 
				   outf, strerror(errno));
			fclose(infp);
			exit(1);
		}
	}
	if (outfp)
		nc = convert_landolt(infp, outfp);
	else 
		nc = convert_landolt(infp, stdout);
	if (nc < 0) {
		err_printf("Error converting landolt table\n");
		fclose(infp);
		if (outfp)
			fclose(outfp);
		exit(1);
	} else {
		fclose(infp);
		if (outfp) {
			info_printf("%d star(s) written\n", nc);
			fclose(outfp);
		}
		exit (0);
	}
}

static void add_image_file_to_list(struct image_file_list *imfl, char *filename)
{
	struct image_file *imf;

	imf = image_file_new();
	imf->filename = strdup(filename);
	imfl->imlist = g_list_append(imfl->imlist, imf);
}


int main(int ac, char **av)
{
	int ret, i;
	char *file;
	struct ccd_frame *fr = NULL;
	char repf[1024]=""; /* report file to be converted */
	char rf[1024]=""; /* recipy file to load */
	char ldf[1024]=""; /* landolt table file to be converted */
	char rcf[1024]=""; /* recipy file to be converted */
	char outf[1024]=""; /* outfile */
	char of[1024]=""; /* obsfile */
	int to_pnm = 0;
	int run_phot = 0;
	int convert_recipy = 0;
	int rep_to_table = 0;
	struct ccd_reduce *ccdr = NULL;
	struct image_file_list *imfl = NULL; /* list of frames to load / process */

	char *rc_fn = NULL;
	char oc;
	int interactive = 0; /* user option to run otherwise batch jobs in
			      * interactive mode */
	int batch = 0; /* batch operations have been selected */

	char *endp;
	double v;

	GtkWidget *window;

	char *shortopts = "D:p:hP:vo:id:b:f:B:M:A:O:usa:T:S:nG:";
	struct option longopts[] = {
		{"debug", required_argument, NULL, 'D'},
		{"dark", required_argument, NULL, 'd'},
		{"bias", required_argument, NULL, 'b'},
		{"flat", required_argument, NULL, 'f'},
		{"badpix", required_argument, NULL, 'B'},
		{"add-bias", required_argument, NULL, 'A'},
		{"multiply", required_argument, NULL, 'M'},
		{"gaussin-blur", required_argument, NULL, 'G'},

		{"output", required_argument, NULL, 'o'},
		{"update-file", no_argument, NULL, 'u'},

		{"stack", no_argument, NULL, 's'},
		{"superflat", no_argument, NULL, 'F'},
		{"align", required_argument, NULL, 'a'},

		{"set", required_argument, NULL, 'S'},

		{"rcfile", required_argument, NULL, 'r'},
		{"obsfile", required_argument, NULL, 'O'},
		{"recipy", required_argument, NULL, 'p'},
		{"convert-rcp", required_argument, NULL, '2'},
		{"to-pnm", no_argument, NULL, 'n'},
		{"convert-landolt", required_argument, NULL, 'L'},
		{"rep-to-table", required_argument, NULL, 'T'},
		{"phot-run", required_argument, NULL, 'P'},
		{"interactive", no_argument, NULL, 'i'},
		{"help", no_argument, NULL, 'h'},
		{"help-all", no_argument, NULL, '3'},
		{"version", no_argument, NULL, '1'},
		{NULL, 0, NULL, 0}
	};

	init_ptable();
	load_params_rc();
	gtk_init (&ac, &av);

	debug_level = 0;

	while ((oc = getopt_long(ac, av, shortopts, longopts, NULL)) > 0) {
		switch(oc) {
		case 'D':
			sscanf(optarg, "%d", &debug_level);
			break;
		case 'h':
			show_usage();
			exit(0);
		case '3':
			help_all();
			exit(0);
		case '1':
			info_printf("%s\n", VERSION);
			exit(0);
		case ':':
			err_printf("option -%c requires an argument\n", optopt);
			exit(1);
		case '?':
			err_printf("Try `gcx -h' for more information\n");
			exit(1);
		case 'i':
			interactive = 1;
			break;
		case 'n':
			to_pnm = 1;
			batch = 1;
			break;
		case 'O':
			strncpy(of, optarg, 1023);
			break;
		case 'o':
			strncpy(outf, optarg, 1023);
			break;
		case 'p':
			sscanf(optarg, "%800s", rf);
			break;
		case 'P':
			sscanf(optarg, "%800s", rf);
			run_phot = 1;
			batch = 1;
			break;
		case '2':
			if (optarg) {
				if (optarg[0] == '-' && optarg[1] == 0)
					;
				else
					strncpy(rcf, optarg, 1023);
			}
			convert_recipy = 1;
			batch = 1;
			break;
		case 'S':
			if (optarg) {
				if (set_par_by_name(optarg))
					err_printf("error setting %s\n", optarg);
			}
			break;
		case 'T':
			if (optarg) {
				if (optarg[0] == '-' && optarg[1] == 0)
					;
				else
					strncpy(repf, optarg, 1023);
			}
			rep_to_table = 1;
			batch = 1;
			break;
		case 'L':
			d3_printf("ldf\n");
			if (optarg)
				strncpy(ldf, optarg, 1023);
			break;


		case 'u':
			if (ccdr == NULL)
				ccdr = ccd_reduce_new();
			ccdr->ops |= IMG_OP_INPLACE;
			break;
		case 's':
			if (ccdr == NULL)
				ccdr = ccd_reduce_new();
			ccdr->ops |= IMG_OP_STACK;
			batch = 1;
			break;
		case 'F':
			if (ccdr == NULL)
				ccdr = ccd_reduce_new();
			ccdr->ops |= IMG_OP_STACK;
			ccdr->ops |= IMG_OP_BG_ALIGN_MUL;
			batch = 1;
			break;
		case 'a':
			if (ccdr == NULL)
				ccdr = ccd_reduce_new();
			ccdr->ops |= IMG_OP_ALIGN;
			ccdr->alignref = image_file_new();
			ccdr->alignref->filename = strdup(optarg);
			batch = 1;
			break;
		case 'b':
			if (ccdr == NULL)
				ccdr = ccd_reduce_new();
			ccdr->bias = image_file_new();
			ccdr->bias->filename = strdup(optarg);
			ccdr->ops |= IMG_OP_BIAS;
			batch = 1;
			break;
		case 'f':
			if (ccdr == NULL)
				ccdr = ccd_reduce_new();
			ccdr->flat = image_file_new();
			ccdr->flat->filename = strdup(optarg);
			ccdr->ops |= IMG_OP_FLAT;
			batch = 1;
			break;
		case 'd':
			if (ccdr == NULL)
				ccdr = ccd_reduce_new();
			ccdr->dark = image_file_new();
			ccdr->dark->filename = strdup(optarg);
			ccdr->ops |= IMG_OP_DARK;
			batch = 1;
			break;
		case 'G':
			v = strtod(optarg, &endp);
			if (endp != optarg) {		
				if (ccdr == NULL)
					ccdr = ccd_reduce_new();
				ccdr->blurv = v;
				ccdr->ops |= IMG_OP_BLUR;
			}
			batch = 1;
			break;
		case 'M':
			v = strtod(optarg, &endp);
			if (endp != optarg) {		
				if (ccdr == NULL)
					ccdr = ccd_reduce_new();
				ccdr->mulv = v;
				ccdr->ops |= IMG_OP_MUL;
			}
			batch = 1;
			break;
		case 'A':
			v = strtod(optarg, &endp);
			if (endp != optarg) {		
				if (ccdr == NULL)
					ccdr = ccd_reduce_new();
				ccdr->addv = v;
				ccdr->ops |= IMG_OP_ADD;
			}
			batch = 1;
			break;
		}
	}

/* now run the various requested operations */

	if (rc_fn != NULL) {/* load extra rc file */
		load_par_file(rc_fn, PAR_NULL);
	}

	if (convert_recipy) { /* convert recipy */
		recipy_file_convert(rcf, outf); /* does not return */
	}

	if (rep_to_table) { /* convert a report into tabular format */
		report_convert(repf, outf); /* does not return */
	}

	if (ldf[0] != 0) { /* convert landolt table */
		d3_printf("landolt convert\n"); /* does not return */
		landolt_file_convert(ldf, outf);
	}

	if (of[0] != 0) { /* search path for obs file */
		if ((strchr(of, '/') == NULL)) {
			file = find_file_in_path(of, P_STR(FILE_OBS_PATH));
			if (file != NULL) {
				strncpy(of, file, 1023);
				of[1023] = 0;
				free(file);
			}
		}
		d3_printf("of is %s\n", of);
	}

	if (rf[0] != 0) { /* search path for recipy file */
		if ((strchr(of, '/') == NULL)) {
			file = find_file_in_path(rf, P_STR(FILE_RCP_PATH));
			if (file != NULL) {
				strncpy(rf, file, 1023);
				rf[1023] = 0;
				free(file);
			}
		}
		d3_printf("rf is %s\n", rf);
	}

	if (ac > optind) { /* we build the image list */
		imfl = image_file_list_new();
		for (i = optind; i < ac; i++)
			add_image_file_to_list(imfl, av[i]);
	}

	if (ccdr != NULL) {
		if (batch && !interactive) {
			if (outf != NULL && outf[0] != 0) {
				if (imfl == NULL) {
					err_printf("No frames to process, exiting\n");
					ccd_reduce_release(ccdr);
					exit(1);
				}
				d3_printf("outf is |%s|\n", outf);
				ret = batch_reduce_frames(imfl, ccdr, outf);
				ccd_reduce_release(ccdr);
				image_file_list_release(imfl);
				exit(ret);
			} else {
				info_printf("no output file name: going interactive\n");
				interactive = 1;
			}
		} 
	}

	window = create_image_window();
	if (!batch || interactive) {
		gtk_widget_show_all(window);
	}

	set_imfl_ccdr(window, ccdr, imfl);

	if (ac > optind) {
/*		if (ccdr != NULL) {
			fr = reduce_frames_load(imfl, ccdr);
			if (fr == NULL)
				exit (1);
		}
*/
		if (fr == NULL)
			fr = read_gz_fits_file(av[optind], P_STR(FILE_UNCOMPRESS));
		if (fr == NULL) {
			err_printf("cannot read file %s", av[optind]);
			exit(1);
		}

		frame_to_channel(fr, window, "i_channel");
		release_frame(fr);
		if (to_pnm) {
			struct image_channel *channel;
			channel = gtk_object_get_data(GTK_OBJECT(window), "i_channel");
			if (channel == NULL) {
				err_printf("oops - no channel\n");
				if (batch && !interactive)
					exit(1);
			}
			if (outf[0] != 0) {
				ret = channel_to_pnm_file(channel, NULL, outf);
			} else {
				ret = channel_to_pnm_file(channel, NULL, NULL);
			}
			if (ret) {
				err_printf("error writing pnm file\n");
				if (batch && !interactive)
					exit(1);
			}
			if (batch && !interactive)
				exit(0);
		}
		if (rf[0] != 0) {
			FILE *output_file = NULL;
			load_rcp_to_window(window, rf);
			match_field_in_window(window);
			if (run_phot) {
				if (outf[0] != 0) {
					output_file = fopen(outf, "a");
					if (output_file == NULL)
						err_printf("Cannot open file %s for appending\n%s\n", 
							   outf, strerror(errno));
				}
				if (output_file == NULL)
					phot_to_fd(window, stdout);
				else {
					phot_to_fd(window, output_file);
					fclose(output_file);
				}
			}
			if (batch && !interactive) {
				exit(0);

			}
		}
		d3_printf("calling gtk_main\n");
		gtk_main ();
		exit(0);
	}
	d3_printf("calling gtk_main, no frame\n");
	if (of[0] != 0) { /* run obs */
		if (run_obs_file(window, of))
			exit (1);
	}
	gtk_main ();
	exit(0);
}
















